### Added feature: MLLP to HTTPS
### By Tiago Rodrigues
### Sectra Iberia, Aug 2022
### Addapted from https://github.com/rivethealth/mllp-http

import functools
import logging
import os
import requests
import socket
import socketserver
import urllib
from .mllp import read_mllp, write_mllp
from .net import read_socket_bytes
from .version import __version__

logger = logging.getLogger(__name__)


def display_address(address):
    return f"{address[0]}:{address[1]}"


class MllpServerOptions:
    def __init__(self, timeout):
        self.timeout = timeout


class MllpHandler(socketserver.StreamRequestHandler):
    # Class for parsing HL7 data as MLLP Server and send it to a host as a HTTPS Client

    def __init__(self, request, address, server, timeout, https_url, https_options):
        self.https_url = https_url
        self.https_options = https_options
        self.timeout = timeout
        super().__init__(request, address, server)

    def handle(self):
        if self.timeout:
            self.request.settimeout(self.timeout)
        self.request.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 10)

        # Session for HTTPS persistent connection
        session = requests.Session()

        # >> Notes on dealing with SSL:
        # Requests verifies SSL certificates for HTTPS requests.
        # By default, SSL verification is enabled, and Requests will
        # throw a SSLError if it’s unable to verify the certificate
        #
        # The private key to the local certificate must be unencrypted.
        # Currently, Requests does not support using encrypted keys.
        #
        # https://requests.readthedocs.io/en/latest/user/advanced/

        local_address = self.request.getsockname()
        remote_address = self.request.getpeername()

        stream = read_socket_bytes(self.rfile)

        try:
            for message in read_mllp(stream):
                try:
                    logger.info("Message: %s bytes", len(message))
                    print(message)
                    headers = {
                        "Forwarded": f"by={display_address(local_address)};for={display_address(remote_address)};proto=mllp",
                        "User-Agent": f"mllp2http/{__version__}",
                        "X-Forwarded-For": display_address(remote_address),
                        "X-Forwarded-Proto": "mllp",
                    }

                    if os.environ.get("HTTP_AUTHORIZATION"):
                        headers["Authorization"] = os.environ["HTTP_AUTHORIZATION"]
                    if os.environ.get("API_KEY"):
                        headers["X-API-KEY"] = os.environ["API_KEY"]

                    if self.https_options.content_type is not None:
                        headers["Content-Type"] = self.https_options.content_type

                    # Disabled the warning because it was causing the program to freeze
                    if not self.https_options.verify:
                        print("Warning! Verify SSL: " + str(self.https_options.verify))
                        requests.packages.urllib3.disable_warnings()

                    # Sending the HL7 data by HTTPS by POST Method
                    response = session.post(
                        urllib.parse.urlunparse(self.https_url),
                        data=message,
                        headers=headers,
                        timeout=self.https_options.timeout,
                        verify=self.https_options.verify,  # To verify server SSL/TLS certificate. Keep as False only
                        # when using a self-sign certificate since the library does not allow self-signed certificates
                    )
                    response.raise_for_status()  # Get ACK response
                except requests.exceptions.HTTPError as e:
                    logger.error("HTTPS response error: %s", e.response.status_code)
                    break
                except Exception as e:
                    logger.error("HTTPS connection error: %s", e)
                    break
                else:
                    content = response.content
                    print(content.decode())
                    logger.info("Response: %s bytes", len(content))
                    write_mllp(self.wfile, content)
                    self.wfile.flush()
        except ConnectionResetError as e:
            # An enception was raised when the MLLP client closed the connection on their side
            logger.info("MLLP Server Disconnected")
        except Exception as e:
            logger.error("Failed read MLLP message: %s", e)


class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    allow_reuse_address = True


class HttpsClientOptions:
    def __init__(self, content_type, timeout, verify):
        self.content_type = content_type
        self.timeout = timeout
        self.verify = verify


def serve(address, options, https_url, https_options):
    logger = logging.getLogger(__name__)

    # Handler for parsing the data from the MLLP socket
    handler = functools.partial(
        MllpHandler,
        https_url=https_url,
        https_options=https_options,
        timeout=options.timeout or None,
    )

    # MLLP Server/Listener
    server = ThreadedTCPServer(address, handler)

    logger.info("\nListening on %s:%s", address[0], address[1])
    server.serve_forever()



import functools
import http.server
import logging
import socket
import threading
import time
from .mllp import send_mllp

logger = logging.getLogger(__name__)


class MllpClientOptions:
    def __init__(self, keep_alive, max_messages, timeout):
        #self.address = address
        self.keep_alive = keep_alive
        self.max_messages = max_messages
        self.timeout = timeout


class MllpClient:
    def __init__(self, address, options):
        self.address = address
        self.options = options
        self.connections = []
        self.lock = threading.Lock()

    def _check_connection(self, connection):
        while not connection.closed:
            elasped = (
                connection.last_update - time.monotonic()
                if connection.last_update is not None
                else 0
            )
            remaining = self.options.keep_alive + elasped
            if 0 < remaining:
                time.sleep(remaining)
            else:
                try:
                    with self.lock:
                        self.connections.remove(connection)
                except ValueError:
                    pass
                else:
                    if self.options.keep_alive > 0:
                        # To keep the connection alive in case keep_alive is -1
                        connection.close()

    def _connect(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if self.options.timeout:
            s.settimeout(self.options.timeout)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 10)
        #print(self.address)
        s.connect(self.address)
        connection = MllpConnection(s)
        if self.options.keep_alive is not None:
            thread = threading.Thread(
                daemon=False,
                target=self._check_connection,
                args=(connection, )
            )
            thread.start()
        return connection

    def send(self, data):
        with self.lock:
            try:
                connection = self.connections.pop()
            except IndexError:
                connection = None
            else:
                connection.last_update = None
        if connection is None:
            connection = self._connect()
        response = connection.send(data)
        if self.options.max_messages <= connection.message_count and self.options.max_messages >= 0:
            connection.close()
        else:
            connection.last_update = time.monotonic()
            with self.lock:
                self.connections.append(connection)
        return response


class MllpConnection:
    def __init__(self, socket):
        self.closed = False
        self.last_update = None
        self.message_count = 0
        self.socket = socket
        #self.responses = read_mllp(read_socket_bytes(self.socket))
        #self.responses = read_mllp(self.socket.recv(1024))
        #print(self.responses)

    def close(self):
        self.close = True
        self.socket.shutdown(2)
        self.socket.close()
        print("Disconnected from MLLP Server")

    def send(self, data):
        #write_mllp(read_socket_bytes(self.socket), data)
        #self.socket.flush()
        self.message_count += 1
        #return next(self.responses)
        return send_mllp(self.socket, data)


# class MllpConnection:
#     def __init__(self, socket):
#         #self.address = address
#         self.cancel = None
#         self.closed = False
#         self.socket = socket
#         self.responses = read_mllp(read_socket_bytes(self.socket))
#
#     def close(self):
#         self.closed = True
#         self.socket.close()
#
#     def send(self, data):
#         write_mllp(read_socket_bytes(self.socket), data)
#         self.socket.flush()
#         self.message_count += 1
#         return next(self.responses)


class HttpServerOptions:
    def __init__(self, timeout, content_type, keep_alive):
        self.timeout = timeout
        self.content_type = content_type
        self.keep_alive = keep_alive


class HttpHandler(http.server.BaseHTTPRequestHandler):
    def __init__(self, request, address, server, mllp_client, content_type, timeout, keep_alive):
        self.content_type = content_type
        self.mllp_client = mllp_client
        self.timeout = timeout
        self.keep_alive = keep_alive
        super().__init__(request, address, server)


    def do_POST(self):
        content_length = int(self.headers["Content-Length"])
        data = self.rfile.read(content_length)
        logger.info("Message: %s bytes", len(data))
        print("Received Data:\n{}".format(data))
        response = self.mllp_client.send(data)
        logger.info("Response: %s bytes", len(response))
        self.send_response(201)
        self.send_header("Content-Length", len(response))
        if self.content_type:
            self.send_header("Content-Type", self.content_type)
        if self.keep_alive is not None:
            self.send_header("Keep-Alive", f"timeout={self.keep_alive}")
        self.end_headers()
        self.wfile.write(response)


def serve(address, options, mllp_address, mllp_options):
    client = MllpClient(mllp_address, mllp_options)

    handler = functools.partial(
        HttpHandler,
        content_type=options.content_type,
        keep_alive=options.keep_alive,
        timeout=options.timeout or None,
        mllp_client=client,
    )

    server = http.server.ThreadingHTTPServer(address, handler)
    logger.info("\nListening on %s:%s", address[0], address[1])
    server.protocol_version = "HTTP/1.1"
    server.serve_forever()



#!/usr/bin/env python3
import os
import setuptools

version = {}
with open("mllp_http_https/version.py", "r") as f:
    exec(f.read(), version)

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    author="Tiago Rodrigues/SECTRA Iberia",
    author_email="tiagoepr@hotmail.com",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    license="MIT",
    description="Translate between MLLP and HTTP/HTTPS",
    entry_points={
        "console_scripts": [
            "mllp2http=mllp_http_https.main:mllp2http",
            "http2mllp=mllp_http_https.main:http2mllp",
            "mllp2https=mllp_http_https.main:mllp2https",
            "https2mllp=mllp_http_https.main:https2mllp",
        ]
    },
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "requests",
        "ssl",
        "certifi",
        "urllib3",
    ],
    name="mllp-https",
    packages=setuptools.find_packages(),
    project_urls={
        "Issues": "https://github.com/tiagoepr/mllp-https/issues",
        "Original Project by Rivet Health": "https://github.com/rivethealth/mllp-http/",
    },
    url="https://github.com/tiagoepr/mllp-https",
    version=version["__version__"],
)



import argparse
import logging
import urllib.parse
from .version import __version__


class ArgumentFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter
):
    pass


def log_level(arg):
    if arg == "error":
        return logging.ERROR
    elif arg == "warn":
        return logging.WARNING
    elif arg == "info":
        return logging.INFO


def url_type(arg):
    return urllib.parse.urlparse(arg)


def http2mllp():
    parser = argparse.ArgumentParser(
        "http2mllp",
        description="""
            HTTP server that proxies an MLLP server.
            Expects an MLLP response message and uses it as the HTTP response.
        """,
        formatter_class=ArgumentFormatter,
    )
    parser.add_argument(
        "-H",
        "--host",
        default="0.0.0.0",
        help="HTTP host",
    )
    parser.add_argument(
        "-p",
        "--port",
        default=8000,
        type=int,
        help="HTTP port",
    )
    parser.add_argument(
        "--keep-alive",
        type=int,
        default=-1,
        help="keep-alive in milliseconds, or unlimited if -1.",
    )
    parser.add_argument(
        "--log-level",
        choices=("error", "warn", "info"),
        default="info",
    )
    parser.add_argument(
        "--mllp-keep-alive",
        type=int,
        default=-1,
        help="keep-alive in milliseconds, or unlimited if -1.",
    )
    parser.add_argument(
        "--mllp-max-messages",
        type=int,
        default=-1,
        help="maximum number of messages per connection, or unlimited if -1.",
    )
    parser.add_argument(
        "--mllp-release",
        default="1",
        choices=("1"),
        help="MLLP release version",
    )
    parser.add_argument(
        "--timeout",
        default=0,
        type=float,
        help="socket timeout, in milliseconds, or unlimited if 0.",
    )
    parser.add_argument(
        "--content-type",
        default="x-application/hl7-v2+er7",
        help="HTTP Content-Type header",
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s {}".format(__version__)
    )
    parser.add_argument(
        "mllp_url",
        #type=url_type,
        help="MLLP URL, Defaulf: hostname"
    )
    parser.add_argument(
        "--mllp_port",
        default=2575,
        type=int,
        help="MLLP PORT",
    )

    args = parser.parse_args()

    import mllp_http_https.http2mllp

    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s %(message)s",
        level=log_level(args.log_level),
    )

    http_server_options = mllp_http_https.http2mllp.HttpServerOptions(
        timeout=args.timeout / 1000,
        content_type=args.content_type,
        keep_alive=args.keep_alive,
    )
    mllp_client_options = mllp_http_https.http2mllp.MllpClientOptions(
        keep_alive=args.mllp_keep_alive / 1000,
        max_messages=args.mllp_max_messages,
        timeout=args.timeout / 100,
    )

    try:
        mllp_http_https.http2mllp.serve(
            address=(
                args.host,
                args.port,
            ),
            options=http_server_options,
            mllp_address=(args.mllp_url, args.mllp_port),
            mllp_options=mllp_client_options,
        )
    except KeyboardInterrupt:
        pass


def mllp2http():
    parser = argparse.ArgumentParser(
        "mllp2http",
        description="MLLP server that proxies an HTTP server. Sends back the HTTP response.",
        formatter_class=ArgumentFormatter,
        epilog="""
environment variables:
    HTTP_AUTHORIZATION - HTTP Authorization header
    X-API-KEY - HTTP X-API-KEY header
        """,
    )
    parser.add_argument(
        "-H",
        "--host",
        default="0.0.0.0",
        help="MLLP host",
    )
    parser.add_argument(
        "-p",
        "--port",
        default=2575,
        type=int,
        help="MLLP port",
    )
    parser.add_argument(
        "--content-type",
        default="x-application/hl7-v2+er7",
        help="HTTP Content-Type header",
    )
    parser.add_argument(
        "--log-level",
        choices=("error", "warn", "info"),
        default="info",
    )
    parser.add_argument(
        "--mllp-release",
        default="1",
        choices=("1"),
        help="MLLP release version",
    )
    parser.add_argument(
        "--timeout",
        default=0,
        type=float,
        help="timeout in milliseconds",
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s {}".format(__version__)
    )
    parser.add_argument("http_url", help="HTTP URL", type=url_type)
    args = parser.parse_args()

    import mllp_http_https.mllp2http

    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s %(message)s",
        level=log_level(args.log_level),
    )

    http_client_options = mllp_http_https.mllp2http.HttpClientOptions(
        content_type=args.content_type,
        timeout=args.timeout if args.timeout else None,
    )
    mllp_server_options = mllp_http_https.mllp2http.MllpServerOptions(
        timeout=args.timeout / 1000 if args.timeout else None
    )

    try:
        mllp_http_https.mllp2http.serve(
            address=(args.host, args.port),
            http_options=http_client_options,
            http_url=args.http_url,
            options=mllp_server_options,
        )
    except KeyboardInterrupt:
        pass


### Added features: MLLP to HTTPS and vice-versa
### By Tiago Rodrigues
### Sectra Iberia, Aug 2022

def mllp2https():
    parser = argparse.ArgumentParser(
        "mllp2https",
        description="MLLP server that proxies an HTTPS server. Sends back the HTTPS response.",
        formatter_class=ArgumentFormatter,
        epilog="""
environment variables:
    HTTP_AUTHORIZATION - HTTP Authorization header
    X-API-KEY - HTTP X-API-KEY header
        """,
    )
    parser.add_argument(
        "-H",
        "--host",
        default="0.0.0.0",
        help="MLLP host",
    )
    parser.add_argument(
        "-p",
        "--port",
        default=2575,
        type=int,
        help="MLLP port",
    )
    parser.add_argument(
        "--content-type",
        default="x-application/hl7-v2+er7",
        help="HTTPS Content-Type header",
    )
    parser.add_argument(
        "--log-level",
        choices=("error", "warn", "info"),
        default="info",
    )
    parser.add_argument(
        "--mllp-release",
        default="1",
        choices=("1"),
        help="MLLP release version",
    )
    parser.add_argument(
        "--timeout",
        default=0,
        type=float,
        help="timeout in milliseconds",
    )
    parser.add_argument(
        "--verify",
        choices=("False", "True"),
        default="True",
        type=str,
        help="Verify SSL certificate on server side. True as default",
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s {}".format(__version__)
    )
    parser.add_argument("https_url", help="HTTPS URL", type=url_type)
    args = parser.parse_args()

    import mllp_http_https.mllp2https

    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s %(message)s",
        level=log_level(args.log_level),
    )

    https_client_options = mllp_http_https.mllp2https.HttpsClientOptions(
        content_type=args.content_type,
        timeout=args.timeout if args.timeout else None,
        verify=True if args.verify == "True" else False,
    )
    mllp_server_options = mllp_http_https.mllp2https.MllpServerOptions(
        timeout=args.timeout / 1000 if args.timeout else None
    )

    try:
        mllp_http_https.mllp2https.serve(
            address=(args.host, args.port),
            https_options=https_client_options,
            https_url=args.https_url,
            options=mllp_server_options,
        )
    except KeyboardInterrupt:
        pass


def https2mllp():
    parser = argparse.ArgumentParser(
        "https2mllp",
        description="""
            HTTPS server that proxies an MLLP server.
            Expects an MLLP response message and uses it as the HTTPS response.
        """,
        formatter_class=ArgumentFormatter,
    )
    parser.add_argument(
        "-H",
        "--host",
        default="0.0.0.0",
        help="HTTPS host",
    )
    parser.add_argument(
        "-p",
        "--port",
        default=8000,
        type=int,
        help="HTTPS port",
    )
    parser.add_argument(
        "--keep-alive",
        type=int,
        default=-1,
        help="keep-alive in milliseconds, or unlimited if -1.",
    )
    parser.add_argument(
        "--log-level",
        choices=("error", "warn", "info"),
        default="info",
    )
    parser.add_argument(
        "--mllp-keep-alive",
        type=int,
        #default=10 * 1000,
        default=-1,
        help="keep-alive in milliseconds, or unlimited if -1.",
    )
    parser.add_argument(
        "--mllp-max-messages",
        type=int,
        default=-1,
        help="maximum number of messages per connection, or unlimited if -1.",
    )
    parser.add_argument(
        "--mllp-release",
        default="1",
        choices=("1"),
        help="MLLP release version",
    )
    parser.add_argument(
        "--timeout",
        default=0,
        type=float,
        help="socket timeout, in milliseconds, or unlimited if 0.",
    )
    parser.add_argument(
        "--content-type",
        default="x-application/hl7-v2+er7",
        help="HTTPS Content-Type header",
    )
    parser.add_argument(
        "mllp_url",
        # type=url_type,
        help="MLLP URL, Defaulf: hostname"
    )
    parser.add_argument(
        "--mllp_port",
        default=2575,
        type=int,
        help="MLLP PORT",
    )
    parser.add_argument(
        "--certfile",
        default="C:/ssl/certfile.crt",
        type=str,
        help="Path for HTTPS Server's SSL/TLS Certificate. Default: C:/ssl/certfile.crt",
    )
    parser.add_argument(
        "--keyfile",
        default="C:/ssl/keyfile.key",
        type=str,
        help="Path for HTTPS Server's SSL/TLS Private Key. Default: C:/ssl/keyfile.key",
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s {}".format(__version__)
    )

    args = parser.parse_args()

    import mllp_http_https.https2mllp

    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(name)s %(message)s",
        level=log_level(args.log_level),
    )

    https_server_options = mllp_http_https.https2mllp.HttpsServerOptions(
        timeout=args.timeout / 1000,
        content_type=args.content_type,
        certfile=args.certfile,
        keyfile=args.keyfile,
        keep_alive=args.keep_alive,
    )
    mllp_client_options = mllp_http_https.https2mllp.MllpClientOptions(
        keep_alive=args.mllp_keep_alive / 1000,
        max_messages=args.mllp_max_messages,
        timeout=args.timeout / 100,
    )

    try:
        mllp_http_https.https2mllp.serve(
            address=(
                args.host,
                args.port,
            ),
            options=https_server_options,
            mllp_address=(args.mllp_url, args.mllp_port),
            mllp_options=mllp_client_options,
        )
    except KeyboardInterrupt:
        pass





# For debugging reasons
#https2mllp()